# SPDX-FileCopyrightText: 2017-2019 Free Software Foundation Europe e.V.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Module that contains reports about files and projects for linting."""

import datetime
import logging
from gettext import gettext as _
from hashlib import md5
from io import StringIO
from os import PathLike
from pathlib import Path
from typing import Iterable, List, NamedTuple, Set
from uuid import uuid4

from . import __version__
from ._util import _LICENSING, _checksum
from .project import Project

_LOGGER = logging.getLogger(__name__)

FileReportInfo = NamedTuple(
    "FileReportInfo",
    [
        ("file_report", "FileReport"),
        ("bad_licenses", Set[str]),
        ("missing_licenses", Set[str]),
    ],
)


class ProjectReport:  # pylint: disable=too-many-instance-attributes
    """Object that holds linting report about the project."""

    def __init__(self):
        self.path = None
        self.licenses = dict()
        self.missing_licenses = dict()
        self.bad_licenses = dict()
        self.read_errors = set()
        self.file_reports = set()

        self._unused_licenses = None
        self._files_without_licenses = None
        self._files_without_copyright = None

    def to_dict(self):
        """Turn the report into a json-like dictionary."""
        return {
            "path": str(Path(self.path).resolve()),
            "licenses": {
                identifier: str(path)
                for identifier, path in self.licenses.items()
            },
            "missing_licenses": {
                lic: [str(file_) for file_ in files]
                for lic, files in self.missing_licenses.items()
            },
            "bad_licenses": {
                lic: [str(file_) for file_ in files]
                for lic, files in self.bad_licenses.items()
            },
            "read_errors": list(map(str, self.read_errors)),
            "file_reports": [report.to_dict() for report in self.file_reports],
        }

    def bill_of_materials(self) -> str:
        """Generate a bill of materials from the project.

        See https://spdx.org/specifications.
        """
        out = StringIO()
        # Write mandatory tags
        out.write("SPDXVersion: SPDX-2.1\n")
        out.write("DataLicense: CC0-1.0\n")
        out.write("SPDXID: SPDXRef-DOCUMENT\n")

        out.write("DocumentName: {}\n".format(Path(self.path).resolve().name))
        # TODO: Generate UUID from git revision maybe
        # TODO: Fix the URL
        out.write(
            "DocumentNamespace: "
            "http://spdx.org/spdxdocs/spdx-v2.1-{}\n".format(uuid4())
        )

        # Author
        # TODO: Fix Person and Organization
        out.write("Creator: Person: Anonymous ()\n")
        out.write("Creator: Organization: Anonymous ()\n")
        out.write("Creator: Tool: reuse-{}\n".format(__version__))

        now = datetime.datetime.utcnow()
        now = now.replace(microsecond=0)
        out.write("Created: {}Z\n".format(now.isoformat()))
        out.write(
            "CreatorComment: <text>This document was created automatically "
            "using available reuse information consistent with "
            "REUSE.</text>\n"
        )

        reports = sorted(self.file_reports, key=lambda x: x.spdxfile.name)

        for report in reports:
            out.write(
                "Relationship: SPDXRef-DOCUMENT describes {}\n".format(
                    report.spdxfile.spdx_id
                )
            )

        for report in reports:
            out.write("\n")
            out.write("FileName: {}\n".format(report.spdxfile.name))
            out.write("SPDXID: {}\n".format(report.spdxfile.spdx_id))
            out.write(
                "FileChecksum: SHA1: {}\n".format(report.spdxfile.chk_sum)
            )
            # IMPORTANT: Make no assertion about concluded license. This tool
            # cannot, with full certainty, determine the license of a file.
            out.write("LicenseConcluded: NOASSERTION\n")

            for lic in sorted(report.spdxfile.licenses_in_file):
                out.write("LicenseInfoInFile: {}\n".format(lic))
            if report.spdxfile.copyright:
                out.write(
                    "FileCopyrightText: <text>{}</text>\n".format(
                        report.spdxfile.copyright
                    )
                )
            else:
                out.write("FileCopyrightText: NONE\n")

        # Licenses
        for lic, path in sorted(self.licenses.items()):
            if lic.startswith("LicenseRef-"):
                out.write("\n")
                out.write("LicenseID: {}\n".format(lic))
                out.write("LicenseName: NOASSERTION\n")

                with (Path(self.path) / path).open() as fp:
                    out.write(
                        "ExtractedText: <text>{}</text>\n".format(fp.read())
                    )

        return out.getvalue()

    @classmethod
    def generate(
        cls, project: Project, paths: Iterable[PathLike] = None
    ) -> "ProjectReport":
        """Generate a ProjectReport from a Project."""
        if paths is None:
            paths = [project.root]

        project_report = cls()
        project_report.path = project.root
        project_report.licenses = project.licenses
        for path in paths:
            for file_ in project.all_files(path):
                try:
                    lint_file_info = FileReport.generate(project, file_)
                except (OSError, UnicodeError):
                    # Translators: %s is a path.
                    _LOGGER.info(_("Could not read %s"), file_)
                    project_report.read_errors.add(file_)
                    continue

                # File report.
                project_report.file_reports.add(lint_file_info.file_report)

                # Bad and missing licenses.
                for license in lint_file_info.missing_licenses:
                    project_report.missing_licenses.setdefault(
                        license, set()
                    ).add(lint_file_info.file_report.path)
                for license in lint_file_info.bad_licenses:
                    project_report.bad_licenses.setdefault(license, set()).add(
                        lint_file_info.file_report.path
                    )

        # More bad licenses
        for name, path in project.licenses.items():
            if name not in project.license_map:
                project_report.bad_licenses.setdefault(name, set()).add(path)

        return project_report

    @property
    def used_licenses(self) -> Set[str]:
        """Set of license identifiers that are found in file reports."""
        return set(self.licenses) - self.unused_licenses

    @property
    def unused_licenses(self) -> Set[str]:
        """Set of license identifiers that are not found in any file report."""
        if self._unused_licenses is not None:
            return self._unused_licenses

        used_licenses = set()
        unused_licenses = set()

        for file_report in self.file_reports:
            for lic in file_report.spdxfile.licenses_in_file:
                used_licenses.add(lic)

        for lic in self.licenses:
            if lic not in used_licenses:
                unused_licenses.add(lic)

        self._unused_licenses = unused_licenses
        return unused_licenses

    @property
    def files_without_licenses(self) -> Iterable[PathLike]:
        """Iterable of paths that have no license information."""
        if self._files_without_licenses is not None:
            return self._files_without_licenses

        files_without_licenses = []

        for file_report in self.file_reports:
            if not file_report.spdxfile.licenses_in_file:
                files_without_licenses.append(file_report.path)

        self._files_without_licenses = files_without_licenses
        return files_without_licenses

    @property
    def files_without_copyright(self) -> Iterable[PathLike]:
        """Iterable of paths that have no copyright information."""
        if self._files_without_copyright is not None:
            return self._files_without_copyright

        files_without_copyright = []

        for file_report in self.file_reports:
            if not file_report.spdxfile.copyright:
                files_without_copyright.append(file_report.path)

        self._files_without_copyright = files_without_copyright
        return files_without_copyright


class _File:  # pylint: disable=too-few-public-methods
    """Represent an SPDX file. Sufficiently enough for our purposes, in any
    case.
    """

    def __init__(self, name, spdx_id=None, chk_sum=None):
        self.name: str = name
        self.spdx_id: str = spdx_id
        self.chk_sum: str = chk_sum
        self.licenses_in_file: List[str] = []
        self.copyright: str = None


class FileReport:
    """Object that holds a linting report about a single file. Importantly,
    it also contains SPDX File information in :attr:`spdxfile`.
    """

    def __init__(self, name: PathLike, path: PathLike):
        self.spdxfile = _File(name)
        self.path = Path(path)

    def to_dict(self):
        """Turn the report into a json-like dictionary."""
        return {
            "path": str(Path(self.path).resolve()),
            "name": self.spdxfile.name,
            "spdx_id": self.spdxfile.spdx_id,
            "chk_sum": self.spdxfile.chk_sum,
            "licenses_in_file": [
                lic for lic in self.spdxfile.licenses_in_file
            ],
            "copyright": self.spdxfile.copyright,
        }

    @classmethod
    def generate(cls, project: Project, path: PathLike) -> FileReportInfo:
        """Generate a FileReport from a path in a Project."""
        path = Path(path)
        if not path.is_file():
            raise OSError("{} is not a file".format(path))

        # pylint: disable=protected-access
        relative = project._relative_from_root(path)
        report = cls("./" + str(relative), path)

        bad_licenses = set()
        missing_licenses = set()

        # Checksum and ID
        report.spdxfile.chk_sum = _checksum(path)
        spdx_id = md5()
        spdx_id.update(str(relative).encode("utf-8"))
        spdx_id.update(report.spdxfile.chk_sum.encode("utf-8"))
        report.spdxfile.spdx_id = "SPDXRef-{}".format(spdx_id.hexdigest())

        spdx_info = project.spdx_info_of(path)
        for expression in spdx_info.spdx_expressions:
            for identifier in _LICENSING.license_keys(expression):
                # Bad license
                if identifier not in project.license_map:
                    bad_licenses.add(identifier)
                # Missing license
                elif identifier not in project.licenses:
                    missing_licenses.add(identifier)

                # Add license to report.
                report.spdxfile.licenses_in_file.append(identifier)

        # Copyright text
        report.spdxfile.copyright = "\n".join(
            sorted(spdx_info.copyright_lines)
        )

        return FileReportInfo(report, bad_licenses, missing_licenses)

    def __hash__(self):
        if self.spdxfile.chk_sum is not None:
            return hash(self.spdxfile.name + self.spdxfile.chk_sum)
        return super().__hash__(self)



# SPDX-FileCopyrightText: 2017-2019 Free Software Foundation Europe e.V.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Misc. utilities for reuse."""

# pylint: disable=ungrouped-imports

import logging
import os
import re
import shutil
import subprocess
from argparse import ArgumentTypeError
from gettext import gettext as _
from hashlib import sha1
from os import PathLike
from pathlib import Path
from typing import BinaryIO, List, Optional, Set

from boolean.boolean import Expression, ParseError
from debian.copyright import Copyright
from license_expression import ExpressionError, Licensing

from . import SpdxInfo

GIT_EXE = shutil.which("git")

_LOGGER = logging.getLogger(__name__)
_LICENSING = Licensing()

_END_PATTERN = r"(?:\*/)*(?:-->)*$"
_IDENTIFIER_PATTERN = re.compile(
    r"SPDX" "-License-Identifier: (.*?)" + _END_PATTERN, re.MULTILINE
)
_COPYRIGHT_PATTERNS = [
    re.compile(r"(SPDX" "-FileCopyrightText: .*?)" + _END_PATTERN),
    re.compile(r"(Copyright .*?)" + _END_PATTERN),
    re.compile(r"(© .*?)" + _END_PATTERN),
]

# Amount of bytes that we assume will be big enough to contain the entire
# comment header (including SPDX tags), so that we don't need to read the
# entire file.
_HEADER_BYTES = 4096


def setup_logging(level: int = logging.WARNING) -> None:
    """Configure logging for reuse."""
    # library_logger is the root logger for reuse. We configure logging solely
    # for reuse, not for any other libraries.
    library_logger = logging.getLogger("reuse")
    library_logger.setLevel(level)

    if not library_logger.hasHandlers():
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        library_logger.addHandler(handler)


def execute_command(
    command: List[str], logger: logging.Logger, **kwargs
) -> subprocess.CompletedProcess:
    """Run the given command with subprocess.run. Forward kwargs. Silence
    output into a pipe unless kwargs override it.
    """
    logger.debug("running %s", " ".join(command))

    stdout = kwargs.get("stdout", subprocess.PIPE)
    stderr = kwargs.get("stderr", subprocess.PIPE)

    return subprocess.run(command, stdout=stdout, stderr=stderr, **kwargs)


def find_root() -> Optional[Path]:
    """Try to find the root of the project from CWD. If none is found, return
    None.
    """
    cwd = Path.cwd()
    if in_git_repo(cwd):
        command = [GIT_EXE, "rev-parse", "--show-toplevel"]
        result = execute_command(command, _LOGGER, cwd=cwd)

        if not result.returncode:
            path = result.stdout.decode("utf-8")[:-1]
            return Path(os.path.relpath(path, cwd))
    return None


def find_licenses_directory(root: PathLike = None) -> Optional[Path]:
    """Find the licenses directory from CWD or *root*. In the following order:

    - LICENSES/ in *root*.

    - Current directory if its name is "LICENSES"

    - LICENSES/ in CWD.

    The returned LICENSES/ directory NEED NOT EXIST. It may still need to be
    created.
    """
    if root is None:
        root = find_root()
    cwd = Path.cwd()
    licenses_path = cwd / "LICENSES"

    if root:
        licenses_path = root / "LICENSES"
    elif cwd.name == "LICENSES":
        licenses_path = cwd

    return licenses_path


def in_git_repo(cwd: PathLike = None) -> bool:
    """Is *cwd* inside of a git repository?

    Always return False if git is not installed.
    """
    if cwd is None:
        cwd = Path.cwd()

    if GIT_EXE:
        command = [GIT_EXE, "status"]
        result = execute_command(command, _LOGGER, cwd=cwd)

        return not result.returncode

    return False


def _all_files_ignored_by_git(root: PathLike) -> Set[str]:
    """Return a set of all files ignored by git. If a whole directory is
    ignored, don't return all files inside of it.

    Return an empty list if git is not installed.
    """
    root = Path(root)

    if GIT_EXE:
        command = [
            GIT_EXE,
            "ls-files",
            "--exclude-standard",
            "--ignored",
            "--others",
            "--directory",
            # TODO: This flag is unexpected.  I reported it as a bug in Git.
            # This flag---counter-intuitively---lists untracked directories
            # that contain ignored files.
            "--no-empty-directory",
            # Separate output with \0 instead of \n.
            "-z",
        ]
        result = execute_command(command, _LOGGER, cwd=root)
        all_files = result.stdout.decode("utf-8").split("\0")
        return set(all_files)
    return set()


def decoded_text_from_binary(binary_file: BinaryIO, size: int = None) -> str:
    """Given a binary file object, detect its encoding and return its contents
    as a decoded string. Do not throw any errors if the encoding contains
    errors:  Just replace the false characters.

    If *size* is specified, only read so many bytes.
    """
    rawdata = binary_file.read(size)
    return rawdata.decode("utf-8", errors="replace")


def _determine_license_path(path: PathLike) -> Path:
    """Given a path FILE, return FILE.license if it exists, otherwise return
    FILE.
    """
    path = Path(path)
    license_path = Path("{}.license".format(path))
    if not license_path.exists():
        license_path = path
    return license_path


def _copyright_from_dep5(path: PathLike, copyright: Copyright) -> SpdxInfo:
    """Find the reuse information of *path* in the dep5 Copyright object."""
    result = copyright.find_files_paragraph(Path(path).as_posix())

    if result is None:
        return SpdxInfo(set(), set())

    return SpdxInfo(
        set(map(_LICENSING.parse, [result.license.synopsis])),
        set(map(str.strip, result.copyright.splitlines())),
    )


def extract_spdx_info(text: str) -> None:
    """Extract SPDX information from comments in a string.

    :raises ExpressionError: if an SPDX expression could not be parsed
    """
    expression_matches = set(map(str.strip, _IDENTIFIER_PATTERN.findall(text)))
    expressions = set()
    copyright_matches = set()
    for expression in expression_matches:
        try:
            expressions.add(_LICENSING.parse(expression))
        except (ExpressionError, ParseError):
            _LOGGER.error(_("Could not parse '%s'"), expression)
            raise
    for line in text.splitlines():
        for pattern in _COPYRIGHT_PATTERNS:
            match = pattern.search(line)
            if match is not None:
                copyright_matches.add(match.groups()[0])
                break

    return SpdxInfo(expressions, copyright_matches)


def make_copyright_line(statement: str, year: str = None) -> str:
    """Given a statement, prefix it with ``SPDX-FileCopyrightText:`` if it is
    not already prefixed with some manner of copyright tag.
    """
    if "\n" in statement:
        raise RuntimeError(f"Unexpected newline in '{statement}'")
    for pattern in _COPYRIGHT_PATTERNS:
        match = pattern.search(statement)
        if match is not None:
            return statement
    if year is not None:
        return f"SPDX-FileCopyrightText: {year} {statement}"
    return f"SPDX-FileCopyrightText: {statement}"


def _checksum(path: PathLike) -> str:
    path = Path(path)

    file_sha1 = sha1()
    with path.open("rb") as fp:
        for chunk in iter(lambda: fp.read(128 * file_sha1.block_size), b""):
            file_sha1.update(chunk)

    return file_sha1.hexdigest()


class PathType:  # pylint: disable=too-few-public-methods
    """Factory for creating Paths"""

    def __init__(self, mode="r", force_file=False, force_directory=False):
        if mode in ("r", "w"):
            self._mode = mode
        else:
            raise ValueError("mode='{}' is not valid".format(mode))
        self._force_file = force_file
        self._force_directory = force_directory
        if self._force_file and self._force_directory:
            raise ValueError(
                "'force_file' and 'force_directory' cannot both be True"
            )

    def __call__(self, string):
        path = Path(string)

        try:
            # pylint: disable=no-else-raise
            if self._mode == "r":
                if path.exists() and os.access(path, os.R_OK):
                    if self._force_file and not path.is_file():
                        raise ArgumentTypeError(
                            _("'{}' is not a file").format(path)
                        )
                    if self._force_directory and not path.is_dir():
                        raise ArgumentTypeError(
                            _("'{}' is not a directory").format(path)
                        )
                    return path
                raise ArgumentTypeError(_("can't open '{}'").format(path))
            else:
                if path.is_dir():
                    raise ArgumentTypeError(
                        _("can't write to directory '{}'").format(path)
                    )
                if path.exists() and os.access(path, os.W_OK):
                    return path
                if not path.exists() and os.access(path.parent, os.W_OK):
                    return path
                raise ArgumentTypeError(_("can't write to '{}'").format(path))
        except OSError:
            raise ArgumentTypeError(_("can't read or write '{}'").format(path))


def spdx_identifier(text: str) -> Expression:
    """argparse factory for creating SPDX expressions."""
    try:
        return _LICENSING.parse(text)
    except (ExpressionError, ParseError):
        raise ArgumentTypeError(
            _("'{}' is not a valid SPDX expression, aborting").format(text)
        )



# SPDX-FileCopyrightText: 2017-2019 Free Software Foundation Europe e.V.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""reuse is a tool for compliance with the REUSE recommendations."""

# pylint: disable=ungrouped-imports,too-many-arguments

import contextlib
import datetime
import gettext
import glob
import hashlib
import logging
import os
import re
import sys
from gettext import gettext as _
from pathlib import Path
from typing import (
    BinaryIO,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Union,
)
from uuid import uuid4

from boolean.boolean import Expression
from debian.copyright import Copyright, NotMachineReadableError

_LOGGER = logging.getLogger(__name__)

_PACKAGE_PATH = os.path.dirname(__file__)
_LOCALE_DIR = os.path.join(_PACKAGE_PATH, "locale")

if gettext.find("reuse", localedir=_LOCALE_DIR):
    gettext.bindtextdomain("reuse", _LOCALE_DIR)
    gettext.textdomain("reuse")
    _LOGGER.debug("translations found at %s", _LOCALE_DIR)
else:
    _LOGGER.debug("no translations found at %s", _LOCALE_DIR)

__author__ = "Carmen Bianca Bakker"
__email__ = "carmenbianca@fsfe.org"
__license__ = "GPL-3.0-or-later"
__version__ = "0.5.0"
__REUSE_version__ = "3.0"


_IGNORE_DIR_PATTERNS = [
    re.compile(r"^\.git$"),
    re.compile(r"^LICENSES$"),
    re.compile(r"^\.reuse$"),
]

_IGNORE_FILE_PATTERNS = [
    re.compile(r"^LICENSE"),
    re.compile(r"^COPYING"),
    # ".git" as file happens in submodules
    re.compile(r"^\.git$"),
    re.compile(r"^\.gitkeep$"),
    re.compile(r".*\.license$"),
    re.compile(r".*\.spdx$"),
]

#: Simple structure for holding SPDX information.
#:
#: The two iterables MUST be sets.
SpdxInfo = NamedTuple(
    "SpdxInfo",
    [("spdx_expressions", Set[Expression]), ("copyright_lines", Set[str])],
)


class ReuseException(Exception):
    """Base exception."""


class IdentifierNotFound(ReuseException):
    """Could not find SPDX identifier for license file."""
